package com.sample;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Reader;
import java.util.Iterator;

import org.drools.RuleBase;
import org.drools.RuleBaseConfiguration;
import org.drools.RuleBaseFactory;
import org.drools.StatefulSession;
import org.drools.WorkingMemory;
import org.drools.audit.WorkingMemoryFileLogger;
import org.drools.common.AbstractRuleBase;
import org.drools.compiler.PackageBuilder;
import org.drools.event.AgendaEventListener;
import org.drools.event.DebugAgendaEventListener;
import org.drools.event.DebugRuleFlowEventListener;
import org.drools.event.DebugWorkingMemoryEventListener;
import org.drools.event.DefaultRuleBaseEventListener;
import org.drools.event.RuleBaseEventListener;
import org.drools.event.RuleFlowEventListener;
import org.drools.event.WorkingMemoryEventListener;
import org.drools.rule.Package;


/**
 * This is a sample file to test.... effect of rule base modifications on corresponding WM.
 */
public class DroolsTest5 {

    public static final void main(String[] args) 
    {
        
    		WorkingMemory workingMemory = null;
    		WorkingMemoryFileLogger logger = null;

            File persistedWorkingMemory = new File("InfStorage/WM3.data");
            
            // Create Listeners
            DebugWorkingMemoryEventListener listener = new DebugWorkingMemoryEventListener();
            DebugAgendaEventListener agendaListener = new DebugAgendaEventListener();
            DebugRuleFlowEventListener ruleflow = new DebugRuleFlowEventListener();
            DefaultRuleBaseEventListener rulbase = new DefaultRuleBaseEventListener();
            
            //load up the rulebase
            RuleBase ruleBase = null;
            
            
			try 
			{
				RuleBaseConfiguration conf = new RuleBaseConfiguration();
				conf.setShadowProxy( true );
				conf.setClassLoader(new INFCustomClassLoader());  // To solve shadow facts issue
				ruleBase = RuleBaseFactory.newRuleBase(conf);
				ruleBase.addPackage(readRulePkg1());
			} 
			catch (Exception e1) 
			{
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
            

            if(! persistedWorkingMemory.exists() )
            {
            	// If WorkingMemory does not exist as serialized form, create it.
            	System.out.println("Creating WM...");
            	workingMemory = ruleBase.newStatefulSession();   

            }
            else
            {
               workingMemory = readWM(ruleBase, persistedWorkingMemory);
            }
            
            
            
            // Add Listeners and logger
            workingMemory.addEventListener(listener);
            workingMemory.addEventListener(agendaListener);
            workingMemory.addEventListener(ruleflow);
            workingMemory.addEventListener(rulbase);
            logger = new WorkingMemoryFileLogger(workingMemory); 
            logger.setFileName("log/event");
        
            // Insert Objects
            insertMessageObject(workingMemory,"Hello World", Message.HELLO);
            insertMessageObject(workingMemory, "Good Bye World", Message.GOODBYE);
            
            System.out.println("Firing Rules...");
            
            workingMemory.fireAllRules();  
          
            // Remove Listeners before WM serialization, as they are not serializable.
            workingMemory.removeEventListener(listener);
            workingMemory.removeEventListener(agendaListener);
            workingMemory.removeEventListener(ruleflow);
            workingMemory.removeEventListener(rulbase);
            workingMemory.removeEventListener((WorkingMemoryEventListener)logger);
            workingMemory.removeEventListener((AgendaEventListener)logger);
            workingMemory.removeEventListener((RuleFlowEventListener)logger);
            workingMemory.removeEventListener((RuleBaseEventListener)logger);
            

            /*
             * Serialize WM
             */
            System.out.println("Serializing WM.....");
            saveWM(workingMemory, persistedWorkingMemory);
            logger.writeToDisk();
            
            /*
             * Reload the WM and then update ruleBase
             */
            
            //((AbstractWorkingMemory)workingMemory).
           /* workingMemory = null;
            System.gc();
            System.gc();
            System.gc();*/
            
            /*
             * Issue 1. Rule Base holds referrence to all the WMs it has been used with. Even if I create a WM
             * then serialize that one and reload the same WM, RuleBase holds 2 referrences for both the WMs. 
			 * I have tried couple of options:
             * a. Making workingMemory ref null & then calling gc(). As Rule Base holds a weak ref for WM, that WM supposed to be 
             * gurbage collected after making workingMemory ref null. But its not happening.
             * 
             * b. Calling ((StatefulSession)workingMemory).dispose(); API. 1st call is success means ref to old WM is removed.
             * But 2nd call onwards the API is throwing Null pointer Exception, though in Drools 4.0.3 this API is working 
             * fine.
             * 
             * c. Calling ((AbstractRuleBase)ruleBase).disposeStatefulSession((StatefulSession)workingMemory);
             *    This solves the purpose, but not sure using this API is correct or not.
             *    
             * Note: Solution b & c removes old WM referrence but still rule base modifications are not
             * getting reflected correctly.  
             *      
             */
           
            System.out.println("No of WM Sessions "+ ruleBase.getStatefulSessions().length);
            ((AbstractRuleBase)ruleBase).disposeStatefulSession((StatefulSession)workingMemory);
            
            
            System.out.println("DeSerialize WM....");
            workingMemory = readWM(ruleBase, persistedWorkingMemory);
            System.out.println("No of WM Sessions "+ ruleBase.getStatefulSessions().length);

            //Add Listeners and logger
            workingMemory.addEventListener(listener);
            workingMemory.addEventListener(agendaListener);
            workingMemory.addEventListener(ruleflow);
            workingMemory.addEventListener(rulbase);
           
         //   logger = new WorkingMemoryFileLogger(workingMemory); //throwing exception so ..comented.
         //   logger.setFileName("log/event1");
            
            try 
            {
            	/*
            	 * Issue 2. After I add new Pkg, the new rule gets executed for old data points but for new
            	 * datapoints the newly added rule is not getting executed.
            	 */
            	System.out.println("Adding new Pkg.");
				ruleBase.addPackage(readRulePkg2());
			} 
            catch (Exception e) 
            {
				e.printStackTrace();
			}
            
            
            // Insert Objects
            insertMessageObject(workingMemory,"Hello World", Message.HELLO);
            insertMessageObject(workingMemory, "Good Bye World", Message.GOODBYE);
            
            System.out.println("Firing Rules...");
            
            workingMemory.fireAllRules();  
          
            // Remove Listeners before WM serialization, as they are not serializable.
            workingMemory.removeEventListener(listener);
            workingMemory.removeEventListener(agendaListener);
            workingMemory.removeEventListener(ruleflow);
            workingMemory.removeEventListener(rulbase);
            workingMemory.removeEventListener((WorkingMemoryEventListener)logger);
            workingMemory.removeEventListener((AgendaEventListener)logger);
            workingMemory.removeEventListener((RuleFlowEventListener)logger);
            workingMemory.removeEventListener((RuleBaseEventListener)logger);
            

            /*
             * Serialize WM
             */
            System.out.println("Serializing WM.....");
            saveWM(workingMemory, persistedWorkingMemory);
      //      ((StatefulSession)workingMemory).dispose();  // throws Null Pointer exception
   //         logger.writeToDisk();
            
            
            
            
//          Add Listeners and logger
            workingMemory.addEventListener(listener);
            workingMemory.addEventListener(agendaListener);
            workingMemory.addEventListener(ruleflow);
            workingMemory.addEventListener(rulbase);
           
            try 
            {
            	/*
            	 * Issue 3. Though I am removing a Pkg, still the rule belonging to that pkg is getting executed.
            	 */
            	System.out.println("Removing Pkg 'com.sample' .");
				ruleBase.removePackage("com.sample");
			} 
            catch (Exception e) 
            {
				e.printStackTrace();
			}
            
            
            // Insert Objects
            insertMessageObject(workingMemory,"Hello World", Message.HELLO);
            insertMessageObject(workingMemory, "Good Bye World", Message.GOODBYE);
            
            System.out.println("Firing Rules...");
            
            workingMemory.fireAllRules();  
          
            // Remove Listeners before WM serialization, as they are not serializable.
            workingMemory.removeEventListener(listener);
            workingMemory.removeEventListener(agendaListener);
            workingMemory.removeEventListener(ruleflow);
            workingMemory.removeEventListener(rulbase);
            workingMemory.removeEventListener((WorkingMemoryEventListener)logger);
            workingMemory.removeEventListener((AgendaEventListener)logger);
            workingMemory.removeEventListener((RuleFlowEventListener)logger);
            workingMemory.removeEventListener((RuleBaseEventListener)logger);
            

            /*
             * Serialize WM
             */
            System.out.println("Serializing WM.....");
            saveWM(workingMemory, persistedWorkingMemory);
         
             
            
    }
    
    private static void insertMessageObject(WorkingMemory workingMemory, String msg, int status)
    {
    	 Message message = null;
    	 message = new Message();
         message.setMessage(  msg );
         message.setStatus( status );
         
         workingMemory.insert( message );
    }

    private static void saveWM(WorkingMemory workingMemory, File persistedWorkingMemory)
    {
    	ObjectOutputStream wm_out = null;
        FileOutputStream wm_file = null;
    	try
        {
         
          System.out.println("Saving WM...");
          wm_file = new
          		 FileOutputStream (persistedWorkingMemory);
          wm_out = new
          		 ObjectOutputStream (wm_file);
 
          wm_out.writeObject (workingMemory);
       
        }
        catch (Exception e)
        {
          e.printStackTrace();
        }
        finally
        {

        	try 
        	{
				wm_out.close();
				wm_out = null;
				wm_file.close();
				wm_file = null;
			} 
        	catch (IOException e) 
        	{

				e.printStackTrace();
			}
		    
        }
    }
    
    private static WorkingMemory readWM(RuleBase ruleBase, File persistedWorkingMemory)
    {
    	
    	System.out.println("Reading WM...");
    	WorkingMemory workingMemory = null;
        FileInputStream wm_in = null;

        try
        {
           wm_in = new  FileInputStream (persistedWorkingMemory);
           
           workingMemory = ruleBase.newStatefulSession(wm_in);
          
           System.out.println("Loaded WM: "+workingMemory);
           
           Iterator itr = workingMemory.iterateObjects();
           
           int count = 0;
           while(itr.hasNext())
           {
        	   count++;
        	   System.out.println("Objs in WM: "+ itr.next());
        	   
           }
           System.out.println("Object Count in Deserialized WM: " + count);
           
        }
        catch (Exception e) 
        {
			e.printStackTrace();
		}
        finally
        {
        	try 
        	{
				wm_in.close();
				wm_in = null;
			} 
        	catch (IOException e) 
        	{
				e.printStackTrace();
			}
        }
        return workingMemory;
    }
    
 
    
	private static Package readRulePkg1() throws Exception 
	{
		//read in the source
		FileInputStream file_in = null;
		ObjectInputStream obj_in = null;
		FileOutputStream file_out = null;
		ObjectOutputStream obj_out = null;
		Package pkg = null;
		
		File pkgFile1 = new File("InfStorage/p1.data");
                
        /*
         * If pkg is already serialized ..deserialize it 
         * or create the pkg compiling .drl file and serialize that.
         */
        if(pkgFile1.exists())
        {
        	file_in = new  FileInputStream (pkgFile1);
        	obj_in = new ObjectInputStream(file_in);
        	pkg = new Package();
        	System.out.println("DeSerializing Pkg 1");
        	pkg.readExternal(obj_in);
        	
        	obj_in.close();
        	obj_in = null;
        	file_in.close();
        	file_in = null;
        }
        else
        {
        	Reader source = new InputStreamReader( DroolsTest5.class.getResourceAsStream( "/Sample.drl" ) );
    		PackageBuilder builder = new PackageBuilder();
    		System.out.println("Compiling & creating Pkg..");
    		builder.addPackageFromDrl( source );
    		pkg = builder.getPackage();
    		file_out = new  FileOutputStream (pkgFile1);
    		obj_out = new ObjectOutputStream(file_out);
    		pkg.writeExternal(obj_out);
    		
    		source.close();
    		source = null;
    		obj_out.close();
    		obj_out = null;
    		file_out.close();
    		file_out = null;
    		
        }
        
        return pkg;
	}
	
	private static Package readRulePkg2() throws Exception 
	{
		//read in the source
		FileInputStream file_in = null;
		ObjectInputStream obj_in = null;
		FileOutputStream file_out = null;
		ObjectOutputStream obj_out = null;
		Package pkg = null;
		
		File pkgFile2 = new File("InfStorage/p2.data");
        
        /*
         * If pkg is already serialized ..deserialize it 
         * or create the pkg compiling .drl file and serialize that.
         */
		
        if(pkgFile2.exists())
        {
        	file_in = new  FileInputStream (pkgFile2);
        	obj_in = new ObjectInputStream(file_in);
        	pkg = new Package();
        	System.out.println("DeSerializing Pkg 2");
        	pkg.readExternal(obj_in);
        	
        	obj_in.close();
        	obj_in = null;
        	file_in.close();
        	file_in = null;
        	
        }
        else
        {
        	Reader source = new InputStreamReader( DroolsTest5.class.getResourceAsStream( "/Sample1.drl" ) );
    		PackageBuilder builder = new PackageBuilder();
    		System.out.println("Compiling & creating Pkg..");
    		builder.addPackageFromDrl( source );
    		pkg = builder.getPackage();
    		file_out = new  FileOutputStream (pkgFile2);
    		obj_out = new ObjectOutputStream(file_out);
    		pkg.writeExternal(obj_out);
    		
    		source.close();
    		source = null;
    		obj_out.close();
    		obj_out = null;
    		file_out.close();
    		file_out = null;
    		
        }
       		
		return pkg;
	}
    
}

